--[[
	en: 'Lock steering axle' Specialization for Farming Simulator 17
	en: This script allows you to lock an unlock the steering axle
	
	de: 'Lenkachse sperren' Spezialisierung für den Landwirtschafts Simulator 17
	de: Dieses Script erlaubt dir die Lenkachse zu sperren und zu entsperren.
	
	Author: 	Ifko[nator]
	Date:		04.01.2019
	Version:	2.0
	
	History:	v1.0 @ 07.05.2017 - initial implementation in FS 17
				v1.5 @ 10.05.2017 - add save function and support for the old wheel entry
									if the steering axle is unlocked, it will steer also in backwards direction
				v1.6 @ 13.05.2017 - fix for MP
				v1.7 @ 20.08.2017 - fix for joining players
				v1.8 @ 13.09.2017 - again fix for MP
				v2.0 @ 04.01.2019 - initial implementation in FS 19
]]

LockSteeringAxles = {};
LockSteeringAxles.debugPriority = 0;

local function printError(errorMessage, isWarning, isInfo)
	local prefix = "::ERROR:: ";
	
	if isWarning then
		prefix = "::WARNING:: ";
	elseif isInfo then
		prefix = "::INFO:: ";
	end;
	
	print(prefix .. "from the LockSteeringAxles.lua: " .. tostring(errorMessage));
end;

local function printDebug(debugMessage, priority, addString)
	if LockSteeringAxles.debugPriority >= priority then
		local prefix = "";
		
		if addString then
			prefix = "::DEBUG:: from the LockSteeringAxles.lua: ";
		end;
		
		print(prefix .. tostring(debugMessage));
	end;
end;

function LockSteeringAxles.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(Attachable, specializations);
end;

function LockSteeringAxles.registerEventListeners(vehicleType)
	local functionNames = {
		"onLoad",
		"onUpdate",
		"saveToXMLFile",
		"onWriteStream",
		"onReadStream",
		"onRegisterActionEvents"
	};
	
	for _, functionName in ipairs(functionNames) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, LockSteeringAxles);
	end;
end;

function LockSteeringAxles.registerFunctions(vehicleType)
	local functionNames = {
		"setSteeringAxleActive"
	};
	
	for _, functionName in ipairs(functionNames) do
		SpecializationUtil.registerFunction(vehicleType, functionName, LockSteeringAxles[functionName]);
	end;
end;

function LockSteeringAxles:onLoad(savegame)
	local specLockSteeringAxles = self.spec_lockSteeringAxles;
	local specWheels = self.spec_wheels;
	local specAttachable = self.spec_attachable;
	
	specLockSteeringAxles.foundSteeringAxle = false;
	
	if #specWheels.wheels == 0 then
		return;
	end;
	
	local wheelId = Utils.getNoNil(specWheels.configurations["wheel"], 1);
    local configKey = string.format("vehicle.wheels.wheelConfigurations.wheelConfiguration(%d)", wheelId - 1);
	
	if not hasXMLProperty(self.xmlFile, "vehicle.wheels.wheelConfigurations.wheelConfiguration(0)") then
		--## fallback to the old wheel entry
		
		configKey = "vehicle";
	end;

    local key = configKey .. ".wheels";
	
	local wheelNumber = 0;
	
    while true do
        local wheelKey = key .. string.format(".wheel(%d)", wheelNumber);	
		
		if not hasXMLProperty(self.xmlFile, wheelKey) then
			printDebug("The vehicle: '" .. self.configFileName .. "' has no steering axle in the 'wheels' TAG. Stop searching now.", 1, true);
			
			break;
		end;
		
		local steeringAxleRotMax = Utils.getNoNil(getXMLFloat(self.xmlFile, wheelKey .. ".steeringAxle#rotMax"), 0);
	
		if steeringAxleRotMax ~= 0 then
			printDebug("Found steering axle in TAG number '" .. wheelNumber + 1 .. "' from the 'wheels' TAG. Stop searching now.", 1, true);
			
			specLockSteeringAxles.foundSteeringAxle = true;
			
			specAttachable.steeringAxleUpdateBackwards = true;
			
			break;
		else
			printDebug("Found no steering axle in TAG number '" .. wheelNumber + 1 .. "' from the 'wheels' TAG. Searching now the next TAG number.", 1, true);
		end;
		
		wheelNumber = wheelNumber + 1;
	end;
	
	if specLockSteeringAxles.foundSteeringAxle then
		if savegame ~= nil then
			specLockSteeringAxles.lockSteeringAxle = Utils.getNoNil(getXMLBool(savegame.xmlFile, savegame.key .. ".lockSteeringAxles#lockSteeringAxle"), false);
		else
			specLockSteeringAxles.lockSteeringAxle = false;
		end;
		
		self:setSteeringAxleActive(specLockSteeringAxles.lockSteeringAxle, false);
	end;
end;

function LockSteeringAxles:onWriteStream(streamId, connection)
	if not connection:getIsServer() then
		local specLockSteeringAxles = self.spec_lockSteeringAxles;
		
		if specLockSteeringAxles.foundSteeringAxle then
			streamWriteBool(streamId, specLockSteeringAxles.lockSteeringAxle);
		end;
	end;
end;

function LockSteeringAxles:onReadStream(streamId, connection)
	if connection:getIsServer() then
		local specLockSteeringAxles = self.spec_lockSteeringAxles;
		
		if specLockSteeringAxles.foundSteeringAxle then	
			specLockSteeringAxles.lockSteeringAxle = streamReadBool(streamId);
			
			self:setSteeringAxleActive(specLockSteeringAxles.lockSteeringAxle, true);
		end;
	end;
end;

function LockSteeringAxles:onRegisterActionEvents(isActiveForInput)
	local specLockSteeringAxles = self.spec_lockSteeringAxles;
	
	if self.isClient and specLockSteeringAxles.foundSteeringAxle then        
		self:clearActionEventsTable(specLockSteeringAxles.actionEvents);
        
		if self:getIsActiveForInput(true) then
            local actionEventId;
            
			_, actionEventId = self:addActionEvent(specLockSteeringAxles.actionEvents, InputAction.TOGGLE_LOCK_STEERING_AXLE_BUTTON, self, LockSteeringAxles.toggleSteeringAxleActive, false, true, false, true, nil);
			
			g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL);
			g_inputBinding:setActionEventTextVisibility(actionEventId, true);
			g_inputBinding:setActionEventActive(actionEventId, false);
		end;
	end;
end;

function LockSteeringAxles.toggleSteeringAxleActive(self, actionName, inputValue, callbackState, isAnalog)
	local specLockSteeringAxles = self.spec_lockSteeringAxles;
	
	self:setSteeringAxleActive(not specLockSteeringAxles.lockSteeringAxle, false);
	
	specLockSteeringAxles.forceUpdate = true;
end;

function LockSteeringAxles:onUpdate(dt, isActiveForInput, isSelected)
	local specLockSteeringAxles = self.spec_lockSteeringAxles;
	local specAttachable = self.spec_attachable;

	if specLockSteeringAxles.foundSteeringAxle then
		if self.isClient then	
			specAttachable.updateSteeringAxleAngle = not specLockSteeringAxles.lockSteeringAxle;
			
			local lockSteeringAxleButton = specLockSteeringAxles.actionEvents[InputAction.TOGGLE_LOCK_STEERING_AXLE_BUTTON];
			
			if lockSteeringAxleButton ~= nil and self:getIsActive() then
				local currentText = Vehicle.LOCK_STEERING_AXLE;
				
				g_inputBinding:setActionEventActive(lockSteeringAxleButton.actionEventId, true);
				
				if specLockSteeringAxles.lockSteeringAxle then
					currentText = Vehicle.UNLOCK_STEERING_AXLE;
				end;
				
				g_inputBinding:setActionEventText(lockSteeringAxleButton.actionEventId, currentText);
			end;
		end;
		
		if specLockSteeringAxles.lockSteeringAxle then
			if specAttachable.steeringAxleAngle > 0 then
				specAttachable.steeringAxleAngle = math.max(specAttachable.steeringAxleAngle - specAttachable.steeringAxleAngleSpeed * dt, 0);
			else
				specAttachable.steeringAxleAngle = math.min(specAttachable.steeringAxleAngle + specAttachable.steeringAxleAngleSpeed * dt, 0);
			end;
		end;
	end;
	
	if specAttachable.updateSteeringAxleAngle then
		local baseVehicle = self:getSteeringAxleBaseVehicle();

		if baseVehicle ~= nil and (self.movingDirection >= 0 or specAttachable.steeringAxleUpdateBackwards) then
			local yRot = Utils.getYRotationBetweenNodes(self.steeringAxleNode, baseVehicle.steeringAxleNode);
			
			if math.abs(yRot) > 1.57 then
				if yRot > 0 then
					yRot = yRot - 3.14;
				else
					yRot = yRot + 3.14;
				end;
			end;
			
			local startSpeed = specAttachable.steeringAxleAngleScaleStart;
			local endSpeed = specAttachable.steeringAxleAngleScaleEnd;
			local scale = MathUtil.clamp(1 + (self:getLastSpeed() - startSpeed) * 1.0 / (startSpeed - endSpeed), 0, 1);
			
			specAttachable.steeringAxleTargetAngle = yRot*scale;
		elseif self:getLastSpeed() > 0.2 then
			specAttachable.steeringAxleTargetAngle = 0;
		end;
		
		local dir = MathUtil.sign(specAttachable.steeringAxleTargetAngle - specAttachable.steeringAxleAngle);
		
		if dir == 1 then
			specAttachable.steeringAxleAngle = math.min(specAttachable.steeringAxleAngle + dir * dt * specAttachable.steeringAxleAngleSpeed, specAttachable.steeringAxleTargetAngle);
		else
			specAttachable.steeringAxleAngle = math.max(specAttachable.steeringAxleAngle + dir * dt * specAttachable.steeringAxleAngleSpeed, specAttachable.steeringAxleTargetAngle);
		end;
		
		if specAttachable.steeringAxleTargetNode ~= nil then
			local angle = MathUtil.clamp(specAttachable.steeringAxleAngle, specAttachable.steeringAxleAngleMinRot, specAttachable.steeringAxleAngleMaxRot);
			
			setRotation(specAttachable.steeringAxleTargetNode, 0, angle, 0);
			self:setMovingToolDirty(specAttachable.steeringAxleTargetNode);
		end;

		specLockSteeringAxles.forceUpdate = specAttachable.steeringAxleAngle ~= specAttachable.steeringAxleTargetAngle; --## force update when trailer is not moving (new for FS 19)
    end;
end;

function LockSteeringAxles:saveToXMLFile(xmlFile, key)
	local specLockSteeringAxles = self.spec_lockSteeringAxles;
	
	if specLockSteeringAxles.foundSteeringAxle then
		setXMLBool(xmlFile, key .. "#lockSteeringAxle", specLockSteeringAxles.lockSteeringAxle);
	end;
end;

function LockSteeringAxles:setSteeringAxleActive(lockSteeringAxle, noEventSend)
	local specLockSteeringAxles = self.spec_lockSteeringAxles;
	
	if lockSteeringAxle ~= specLockSteeringAxles.lockSteeringAxle then
		specLockSteeringAxles.lockSteeringAxle = lockSteeringAxle;
		
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(LockSteeringAxlesEvent:new(self, lockSteeringAxle), nil, nil, self);
			else
				g_client:getServerConnection():sendEvent(LockSteeringAxlesEvent:new(self, lockSteeringAxle));
			end;
		end;
	end;
end;

--## MP Stuff

--## turn steering axle on/off

LockSteeringAxlesEvent = {};
LockSteeringAxlesEvent_mt = Class(LockSteeringAxlesEvent, Event);

InitEventClass(LockSteeringAxlesEvent, "LockSteeringAxlesEvent");

function LockSteeringAxlesEvent:emptyNew()
	local self = Event:new(LockSteeringAxlesEvent_mt);
    
	return self;
end;

function LockSteeringAxlesEvent:new(trailer, lockSteeringAxle)
	local self = LockSteeringAxlesEvent:emptyNew();
	
	self.trailer = trailer;
	self.steerAxleIsLocked = lockSteeringAxle;
	
	return self;
end;

function LockSteeringAxlesEvent:readStream(streamId, connection)
	self.trailer = NetworkUtil.readNodeObject(streamId);
	self.steerAxleIsLocked = streamReadBool(streamId);
	
    self:run(connection);
end;

function LockSteeringAxlesEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.trailer);
	
	streamWriteBool(streamId, self.steerAxleIsLocked);
end;

function LockSteeringAxlesEvent:run(connection)
	if not connection:getIsServer() then
		g_server:broadcastEvent(LockSteeringAxlesEvent:new(self.trailer, self.steerAxleIsLocked), nil, connection, self.trailer);
	end;
	
    if self.trailer ~= nil then
        self.trailer:setSteeringAxleActive(self.steerAxleIsLocked, true);
	end;
end;